-- name: Martlet Moveset
-- description: Martlet's moveset.

if (not _G.charSelectExists) then return end

FLYING_DIVE_TIME = 15

FLYING_MIN_GRAVITY = 10.0
FLYING_MAX_GRAVITY = 100.0
FLYING_MAX_STAMINA = 150
FLYING_MAX_TIME = 150
FLYING_SPEED = 45.0
FLYING_MIN_SPEED = 30.0

FLYING_ICON_MIN_POSITION = -64
FLYING_ICON_MAX_POSITION = 16

local TEX_MARTLET_FLYING_ICON = get_texture_info("martlet-flying-icon")

gStateExtras = {}
for i = 0,(MAX_PLAYERS - 1) do
    gStateExtras[i] = {}
    m = gMarioStates[i]
    e = gStateExtras[i]
	
	e.flightFromDive = false
	e.flightDiveTimer = 0
	
	e.flightStamina = FLYING_MAX_STAMINA
	e.flightTimer = 0
	e.flightGravity = 0
	
	e.flightTier = 0
	
	e.flightIconPos = FLYING_ICON_MIN_POSITION
end

local function switch(param, caseTable)
    local case = caseTable[param]
    if case then return case() end
    local def = caseTable['default']
    return def and def() or nil
end

-- AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
local function load_level_init_text(m, arg)
	local p = gNetworkPlayers[0]
	
	local gotAchievement = true
	if (p.currLevelNum == LEVEL_VCUTM) then
		gotAchievement = save_file_get_flags() & SAVE_FLAG_HAVE_VANISH_CAP ~= 0
	elseif (p.currLevelNum == LEVEL_COTMC) then
		gotAchievement = save_file_get_flags() & SAVE_FLAG_HAVE_METAL_CAP ~= 0
	elseif (p.currLevelNum == LEVEL_TOTWC) then
		gotAchievement = save_file_get_flags() & SAVE_FLAG_HAVE_WING_CAP ~= 0
	else
		gotAchievement = save_file_get_star_flags(get_current_save_file_num() - 1, p.currCourseNum - 1) ~= 0
	end
	
	if (not gotAchievement) then
		local dialogID = m.area.dialog[arg]
		if (p.currLevelNum == LEVEL_VCUTM) then dialogID = gBehaviorValues.dialogs.VanishCourseDialog end
		if (p.currLevelNum == LEVEL_COTMC) then dialogID = gBehaviorValues.dialogs.MetalCourseDialog end
		if (p.currLevelNum == LEVEL_TOTWC) then dialogID = gBehaviorValues.dialogs.WingCourseDialog end
		
		--level_set_transition(-1, nil)
		create_dialog_box(dialogID)
	end
end

-- Replace default flying action to one that doesn't strictly require the Wing Cap if Martlet's moveset is enabled
ACT_MARTLET_FLYING = allocate_mario_action(0x099|ACT_FLAG_AIR|ACT_FLAG_DIVING|ACT_FLAG_ATTACKING|ACT_FLAG_SWIMMING_OR_FLYING)
function act_martlet_flying(m)
	local movesetEnabled = (not _G.charSelect.stopMovesets) and (_G.charSelect.get_options_status(_G.charSelect.optionTableRef.localMoveset) ~= 0)
    local startPitch = m.faceAngle.x
	
	-- Mr. Sunshine: SHUT UP!!!!!
	m.flags = m.flags | MARIO_UNKNOWN_18

    if (m.input & INPUT_Z_PRESSED ~= 0) then
        if (m.area.camera.mode == CAMERA_MODE_BEHIND_MARIO) then
            if (m.playerIndex == 0) then
                if (newcam_active == 0) then
                    set_camera_mode(m.area.camera, m.area.camera.defMode, 1)
                else
                    m.area.camera.mode = CAMERA_MODE_NEWCAM
                    gLakituState.mode = CAMERA_MODE_NEWCAM
                end
            end
        end
        return set_mario_action(m, ACT_GROUND_POUND, 1)
    end

	-- Why couldn't I just disable this part if I want bird to fly without hat? Am I stupid?
    if ((not movesetEnabled) and (m.flags & MARIO_WING_CAP == 0)) then
        if (m.area.camera.mode == CAMERA_MODE_BEHIND_MARIO) then
            if (m.playerIndex == 0) then
                if (newcam_active == 0) then
                    set_camera_mode(m.area.camera, m.area.camera.defMode, 1)
                else
                    m.area.camera.mode = CAMERA_MODE_NEWCAM
                    gLakituState.mode = CAMERA_MODE_NEWCAM
                end
            end
        end
        if (m.playerIndex == 0) then return set_mario_action(m, ACT_FREEFALL, 0) end
    end

    if (m.area.camera.mode ~= CAMERA_MODE_BEHIND_MARIO) then
        if (m.playerIndex == 0) then
            if (newcam_active == 0) then
                set_camera_mode(m.area.camera, CAMERA_MODE_BEHIND_MARIO, 1)
            else
                m.area.camera.mode = CAMERA_MODE_NEWCAM
                gLakituState.mode = CAMERA_MODE_NEWCAM
            end
        end
    end

    if (m.actionState == 0) then
        if (m.actionArg == 0) then
            set_character_animation(m, CHAR_ANIM_FLY_FROM_CANNON)
        else
            set_character_animation(m, CHAR_ANIM_FORWARD_SPINNING_FLIP)
            if (m.marioObj.header.gfx.animInfo.animFrame == 1) then
                play_sound(SOUND_ACTION_SPIN, m.marioObj.header.gfx.cameraToObject)
            end
        end

        if (is_anim_at_end(m) ~= 0) then
            if (m.actionArg == 2) then
                if (m == gMarioStates[0]) then
                    load_level_init_text(m, 0)
                end
                m.actionArg = 1
            end

            set_character_animation(m, CHAR_ANIM_WING_CAP_FLY)
            m.actionState = 1
        end
    end

    update_flying(m)
	
	switch(perform_air_step(m, 0), {
		[AIR_STEP_NONE] = function()
			m.marioObj.header.gfx.angle.x = -m.faceAngle.x
			m.marioObj.header.gfx.angle.z = m.faceAngle.z
			m.actionTimer = 0
		end,
		[AIR_STEP_LANDED] = function()
			set_mario_action(m, ACT_DIVE_SLIDE, 0)
			
			set_character_animation(m, CHAR_ANIM_DIVE)
			set_anim_to_frame(m, 7)
			
			m.faceAngle.x = 0
			
			if (m.playerIndex == 0) then
				if (newcam_active == 0) then
					set_camera_mode(m.area.camera, m.area.camera.defMode, 1)
				else
					m.area.camera.mode = CAMERA_MODE_NEWCAM
					gLakituState.mode = CAMERA_MODE_NEWCAM
				end
			end
			queue_rumble_data_mario(m, 5, 60)
		end,
		[AIR_STEP_HIT_WALL] = function()
			if (m.wall ~= nil) then
				mario_set_forward_vel(m, -16)
				m.faceAngle.x = 0
				
				if (m.vel.y > 0) then
					m.vel.y = 0
				end
				
				if (m.flags & MARIO_METAL_CAP ~= 0) then play_sound(SOUND_ACTION_METAL_BONK, m.marioObj.header.gfx.cameraToObject)
				else play_sound(SOUND_ACTION_BONK, m.marioObj.header.gfx.cameraToObject) end
				
				set_mario_particle_flags(m, PARTICLE_VERTICAL_STAR, false)
				set_mario_action(m, ACT_BACKWARD_AIR_KB, 0)
				
				if (m.playerIndex == 0) then
					if (newcam_active == 0) then
						set_camera_mode(m.area.camera, m.area.camera.defMode, 1)
					else
						m.area.camera.mode = CAMERA_MODE_NEWCAM
						gLakituState.mode = CAMERA_MODE_NEWCAM
					end
				end
			else
				if (m.actionTimer == 0) then play_sound(SOUND_ACTION_HIT, m.marioObj.header.gfx.cameraToObject) end
				m.actionTimer = m.actionTimer + 1
				
				if (m.actionTimer == 30) then m.actionTimer = 0 end
				
				m.faceAngle.x = m.faceAngle.x - 0x200
				if (m.faceAngle.x < -0x2AAA) then m.faceAngle.x = -0x2AAA end
				
				m.marioObj.header.gfx.angle.x = -m.faceAngle.x
				m.marioObj.header.gfx.angle.z = m.faceAngle.z
			end
		end,
		[AIR_STEP_HIT_LAVA_WALL] = function()
			lava_boost_on_wall(m)
		end
	})
		
	if (m.faceAngle.x > 0x800 and m.forwardVel >= 48) then
		set_mario_particle_flags(m, PARTICLE_DUST, false)
	end
	
	local isFlyingFast = (startPitch <= 0 and m.faceAngle.x > 0 and m.forwardVel >= 48) and
						 ((m.flags & MARIO_WING_CAP ~= 0) or
						 (m.playerIndex == 0 and movesetEnabled))
	if (isFlyingFast) then
		play_sound(SOUND_ACTION_FLYING_FAST, m.marioObj.header.gfx.cameraToObject)
		play_character_sound(m, CHAR_SOUND_YAHOO_WAHA_YIPPEE)
		queue_rumble_data_mario(m, 50, 40)
	end
	
	play_sound(SOUND_MOVING_FLYING, m.marioObj.header.gfx.cameraToObject)
	adjust_sound_for_speed(m)
	return false
end

-- No gravity while flying! Not allowed!
function act_martlet_flying_gravity(m)
	return true
end

function martlet_update(m)
	local e = gStateExtras[m.playerIndex]
	
	if (m.playerIndex == 0 and m.action == ACT_FLYING) then set_mario_action(m, ACT_MARTLET_FLYING, m.actionArg) end
	
	local hasWingCap = m.flags & MARIO_WING_CAP ~= 0
	local hasVanishCap = m.flags & MARIO_VANISH_CAP ~= 0
	local hasMetalCap = m.flags & MARIO_METAL_CAP ~= 0
	
	--[[if (m.playerIndex == 0) then
		-- Narcolepsy
		if m.action == ACT_IDLE then
			m.actionTimer = m.actionTimer + 1
		end
		-- Delay sleeping while in dialog
		local isInDialog = get_dialog_box_state()
		if isInDialog ~= 0 and m.action == ACT_IDLE then
			m.actionTimer = 0
		end
	end]]--
	
	-- Apply flutter and trigger flight upon cannon launch
	if (m.playerIndex == 0) then
		local applyFlutter = (not (m.action == ACT_TWIRLING and m.vel.y < 0) and
							 (not (m.action == ACT_LONG_JUMP or m.action == ACT_SLIDE_KICK or m.action == ACT_BBH_ENTER_SPIN)) and
							 (not (m.action == ACT_LAVA_BOOST or m.action == ACT_FALL_AFTER_STAR_GRAB)) and
							 (m.action ~= ACT_GETTING_BLOWN) and
							 (not (m.input & INPUT_A_DOWN ~= 0 and m.vel.y > 20 and m.action & ACT_FLAG_CONTROL_JUMP_HEIGHT ~= 0))) and
							 (e.flightTier == 1 and (not hasWingCap) and m.marioBodyState.wingFlutter == 0 and m.vel.y < 0 and m.input & INPUT_A_DOWN ~= 0)
		if (applyFlutter) then m.vel.y = math.max(m.vel.y - 2, -37.5) end
		if ((m.action == ACT_SHOT_FROM_CANNON) and (m.vel.y < 0)) then set_mario_action(m, ACT_FLYING, 0) end -- No Wing Cap necessary!
	end
	
	if m.action ~= ACT_DIVE then
		if (m.playerIndex == 0) then
			e.flightFromDive = false
			e.flightDiveTimer = 0
		end
	end
	
	-- Dive out of ground pounds
	if (e.flightStamina > 0 and m.action == ACT_GROUND_POUND and m.input & INPUT_B_PRESSED ~= 0) then
		if (m.playerIndex == 0) then
			m.forwardVel = 30
			m.faceAngle.y = m.intendedYaw
			m.vel.y = 30
			set_mario_action(m, ACT_DIVE, 0)
			m.particleFlags = m.particleFlags | PARTICLE_DUST
			
			e.flightFromDive = true
			e.flightStamina = math.max(e.flightStamina - 30, 0)
			if (hasWingCap) then e.flightStamina = math.max(e.flightStamina, FLYING_MAX_STAMINA - 60, 30) end
		end
    end
	
	-- Triple Jump out of ground pounds
	if (e.flightStamina > 0 and m.action == ACT_GROUND_POUND_LAND and m.input & INPUT_A_PRESSED ~= 0) then
		if (m.playerIndex == 0) then
			if (not hasWingCap) then set_mario_action(m, ACT_TRIPLE_JUMP, 0)
			else set_mario_action(m, ACT_FLYING_TRIPLE_JUMP, 0) end
			m.particleFlags = m.particleFlags | PARTICLE_HORIZONTAL_STAR | PARTICLE_MIST_CIRCLE
		end
	end
	
	-- Initiate flight from diving out of ground pounds
	if m.action == ACT_DIVE and e.flightFromDive then
		if (m.playerIndex == 0) then e.flightDiveTimer = approach_s32(e.flightDiveTimer, FLYING_DIVE_TIME, 1, 1) end
		if e.flightDiveTimer >= FLYING_DIVE_TIME then
			--set_mario_animation(m, CHAR_ANIM_WING_CAP_FLY)
			if (m.playerIndex == 0) then
				set_mario_action(m, ACT_FLYING, 1)
				
				e.flightFromDive = false
				e.flightDiveTimer = 0
			end
		end
	-- Initiate flight from ANY dive
	elseif (e.flightStamina > 0 and m.action == ACT_DIVE and m.input & INPUT_A_PRESSED ~= 0 and (m.pos.y - m.floorHeight) >= 100) then
		--set_mario_animation(m, CHAR_ANIM_WING_CAP_FLY)
		if (m.playerIndex == 0) then
			set_mario_action(m, ACT_FLYING, 1)
			
			e.flightFromDive = false
			e.flightDiveTimer = 0
			e.flightStamina = math.max(e.flightStamina - 30, 0)
			if (hasWingCap) then e.flightStamina = math.max(e.flightStamina, FLYING_MAX_STAMINA - 60, 30) end
		end
	end
	
	-- Innate flight	
	local isFlying = (m.action == ACT_FLYING or m.action == ACT_MARTLET_FLYING)
	local isGroundPounding = m.action == ACT_GROUND_POUND
	local isDiving = m.action == ACT_DIVE
	
	-- Set flightTier to 0 during certain actions
	local resetFlightTier = (m.action & ACT_FLAG_AIR == 0)
	
	-- Trigger flight upon triple jump, distinguish from regular flying with a Wing Cap using flightTier value
	-- Triggering flight with a triple jump requires you to press A again in midair from a decent height
	if (((m.action == ACT_TRIPLE_JUMP or m.action == ACT_SPECIAL_TRIPLE_JUMP) and (m.pos.y - m.floorHeight) >= 100 and (m.input & INPUT_A_PRESSED ~= 0))) then
		if (m.playerIndex == 0) then set_mario_action(m, ACT_FLYING, 1) end
	end
	
	-- Set flightTier
	if (m.playerIndex == 0) then
		if (isFlying) then
			if (hasWingCap) then e.flightTier = 2
			else e.flightTier = 1 end
		elseif resetFlightTier then e.flightTier = 0 end
	end
	
	if (m.playerIndex == 0 and m.action == ACT_FLYING) then set_mario_action(m, ACT_MARTLET_FLYING, m.actionArg) end
	
	-- Flight characteristics unique to Martlet
	if (m.playerIndex == 0) then
		isFlying = (m.action == ACT_FLYING or m.action == ACT_MARTLET_FLYING)
		if isFlying then			
			if (hasWingCap and m.controller.buttonDown & A_BUTTON ~= 0) then e.isBraking = true
			else e.isBraking = false end
			
			local maxTime = math.max(FLYING_MAX_TIME, 30)
			local minTime = math.min(maxTime, 15)
			local minStamina = math.max(FLYING_MAX_STAMINA - 60, 30)
			local resetPitch = (e.flightTimer <= minTime)
			if resetPitch then m.faceAngle.x = 0 end
					
			-- Powered up flight with Wing Cap
			if hasWingCap then
				local minFlyingSpeed = FLYING_SPEED
				if (e.isBraking) then
					minFlyingSpeed = FLYING_MIN_SPEED
					m.forwardVel = approach_f32_asymptotic(m.forwardVel, FLYING_MIN_SPEED, 0.25)
				end
				--elseif (m.forwardVel < FLYING_SPEED) then m.forwardVel = approach_f32(m.forwardVel, FLYING_SPEED, 5.0, 0.0) end
				
				-- Martlet has a minimum flying speed regardless of pitch
				m.forwardVel = math.max(m.forwardVel, minFlyingSpeed)
				
				-- Visual sparkle effect and pitch correction
				m.particleFlags = m.particleFlags | PARTICLE_SPARKLES
				if (m.controller.stickY == 0) then
					if (e.flightTimer > minTime) and not (e.isBraking and m.forwardVel < FLYING_SPEED) then m.faceAngle.x = m.faceAngle.x - 79
					elseif (m.faceAngle.x < 0) then m.faceAngle.x = approach_s16_asymptotic(m.faceAngle.x, 0, 50) end
				end
				
				e.flightStamina = approach_s32(e.flightStamina, minStamina, 1, 1)
				e.flightTimer = approach_s32(e.flightTimer, maxTime, 1, 1)
				e.flightGravity = 0
			-- Weaker flight without Wing Cap
			else
				-- Apply slight gravity effect during flight, pulling Martlet towards the ground
				-- This effect becomes severe once stamina runs out
				e.flightGravity = math.max(e.flightGravity, FLYING_MIN_GRAVITY)
				if e.flightStamina <= 0 then e.flightGravity = approach_f32_asymptotic(e.flightGravity, FLYING_MAX_GRAVITY, 0.1) end
				--if e.flightTimer >= FLYING_MAX_TIME then e.flightGravity = approach_f32_asymptotic(e.flightGravity, FLYING_MAX_GRAVITY, 0.1) end
				e.flightStamina = approach_s32(e.flightStamina, 0, 1, 1)
				e.flightTimer = approach_s32(e.flightTimer, FLYING_MAX_TIME, 1, 1)
				
				-- Flight speed is capped
				if (m.forwardVel <= FLYING_SPEED or e.flightTimer >= 30) then m.forwardVel = math.min(m.forwardVel, FLYING_SPEED)
				else m.forwardVel = approach_f32_asymptotic(m.forwardVel, FLYING_SPEED, 0.1) end
				
				-- Tilting upwards during flight applies heavy gravity
				local gravNormal = e.flightGravity * 0.25
				local gravTilt = e.flightGravity * 0.75
				
				m.forwardVel = math.max(m.forwardVel, FLYING_MIN_SPEED * math.max(math.min(((6400 - m.faceAngle.x) / 6400), 1), 0))
				m.pos.y = m.pos.y - (gravNormal + (gravTilt * math.max(math.min((m.faceAngle.x / 6400), 1), 0)))
			end
		-- Restore stamina if on the ground, swimming, or launched from a cannon
		else
			if ((m.pos.y - m.floorHeight) <= 0 or (m.action & ACT_FLAG_SWIMMING ~= 0) or (m.action == ACT_SHOT_FROM_CANNON)) then e.flightStamina = FLYING_MAX_STAMINA end
			e.flightTimer = 0
			e.flightGravity = 0
			e.isBraking = false
		end
	end
end

-- The special Triple Jump should ACTUALLY BE SPECIAL GODDAMMIT
function on_set_martlet_action(m)
	if (m.playerIndex ~= 0) then return end
	if (m.action == ACT_FLYING) then set_mario_action(m, ACT_MARTLET_FLYING, m.actionArg) end
	if (m.action == ACT_SPECIAL_TRIPLE_JUMP) then
		set_mario_y_vel_based_on_fspeed(m, 99, 0)
		m.flags = m.flags | MARIO_UNKNOWN_18 	-- Mr. Sunshine: SHUT UP!!!!!
	end
end

-- Grabbing onto poles or overhangs adds one second to the stamina timer
function before_set_martlet_action(m, nextAct)
	if (m.playerIndex ~= 0) then return end
	if not nextAct then return end
	
	if (nextAct == ACT_FLYING) then set_mario_action(m, ACT_MARTLET_FLYING, m.actionArg) end
	
	local e = gStateExtras[m.playerIndex]
	
	local isOnPole = (m.action & ACT_FLAG_ON_POLE == 0) and (nextAct & ACT_FLAG_ON_POLE ~= 0)
	local isHanging = (m.action & ACT_FLAG_HANGING == 0) and (nextAct & ACT_FLAG_HANGING ~= 0)

	if (isOnPole or isHanging) then e.flightStamina = math.min(e.flightStamina + 30, FLYING_MAX_STAMINA) end
end

-- Draw custom HUD
function on_martlet_hud_render()
	local m = gMarioStates[0]
	local e = gStateExtras[0]
	
	-- Display only when needed
	local inMidair = (m.action & ACT_FLAG_AIR ~= 0)
	local isOnPole = (m.action & ACT_FLAG_ON_POLE ~= 0)
	local isHanging = (m.action & ACT_FLAG_HANGING ~= 0)
	
	local isDisplayed = ((inMidair or isOnPole or isHanging) and (e.flightTier > 1 or e.flightStamina < FLYING_MAX_STAMINA))
	if not (isDisplayed) then
		e.flightIconPos = approach_f32_asymptotic(e.flightIconPos, FLYING_ICON_MIN_POSITION, 0.25)
	else
		e.flightIconPos = approach_f32_asymptotic(e.flightIconPos, FLYING_ICON_MAX_POSITION, 0.25)
	end
	
	e.flightIconPos = math.min(e.flightIconPos, FLYING_ICON_MAX_POSITION)
	e.flightIconPos = math.max(e.flightIconPos, FLYING_ICON_MIN_POSITION)
	
	local scale = 1.0
	
	local texWidth = TEX_MARTLET_FLYING_ICON.width * scale
	local texHeight = TEX_MARTLET_FLYING_ICON.height * scale
	
	if (not isDisplayed) and (e.flightIconPos <= -texWidth) then
		e.flightIconPos = FLYING_ICON_MIN_POSITION
		return
	end
	
	djui_hud_set_resolution(RESOLUTION_N64)
    djui_hud_set_font(FONT_RECOLOR_HUD)
    djui_hud_set_color(255, 255, 255, 255)
	
	local text = tostring(math.ceil(e.flightStamina / 30))
	if (m.flags & MARIO_WING_CAP ~= 0) then text = 'M' end
	
	local screenWidth = djui_hud_get_screen_width()
	local screenHeight = djui_hud_get_screen_height()
	
	local length = djui_hud_measure_text(text) * scale
	
    local iconX = e.flightIconPos
	local iconY = screenHeight * 0.25 - (texHeight * 0.5)
	
	local textX = iconX + (length * 0.5)
	local textY = iconY
	
	djui_hud_render_texture(TEX_MARTLET_FLYING_ICON, iconX, iconY, scale, scale)
	
	local r = 0
	local g = 0
	local b = 255
	
	-- Change colors depending on how many seconds are left (the default is pure blue)
	switch(text, {
		['M'] = function() -- Display a rainbow M while using the Wing Cap
			local gTimer = get_global_timer()
			local hue = math.fmod(gTimer, 90) * 4
			hue = math.min(math.max(hue, 0), 360)
			
			r = 0
			g = 0
			b = 0
			
			-- This is clunky as hell but it works
			if (hue <= 120) then r = (120 - hue) * 4.25
			elseif (hue >= 240) then r = (hue - 240) * 4.25 end
			
			if (hue <= 180) then g = (hue * 4.25)
			else g = (240 - hue) * 4.25 end
			
			if (hue >= 180) and (hue <= 300) then b = (hue - 180) * 4.25
			elseif (hue > 300) then b = (360 - hue) * 4.25 end
			
			r = math.min(math.max(r, 0), 255)
			g = math.min(math.max(g, 0), 255)
			b = math.min(math.max(b, 0), 255)			
		end,
		['4'] = function() -- Cyan
			r = 0
			g = 255
			b = 255
		end,
		['3'] = function() -- Green
			r = 0
			g = 255
			b = 0
		end,
		['2'] = function() -- Yellow
			r = 255
			g = 255
			b = 0
		end,
		['1'] = function() -- Orange
			r = 255
			g = 128
			b = 0
		end,
		['0'] = function() -- Red
			r = 255
			g = 0
			b = 0
		end
	})
			
	djui_hud_set_color(r, g, b, 255)
	djui_hud_print_text(text, textX, textY, 1)
end

hook_mario_action(ACT_MARTLET_FLYING, { every_frame = act_martlet_flying, gravity = act_martlet_flying_gravity })